# -*- coding: utf-8 -*-
"""
Sub-package to compute hybrid solar-wind generation stats
"""
from reVX.hybrid_stats.hybrid_stats import HybridStats
from reVX.hybrid_stats.temporal_agg import TemporalAgg
