
# -*- coding: utf-8 -*-
"""
Compute setbacks
"""
from .wind_setbacks import (StructureWindSetbacks, RoadWindSetbacks,
                            RailWindSetbacks, TransmissionWindSetbacks)
from .solar_setbacks import ParcelSetbacks
