# -*- coding: utf-8 -*-
"""
Compute prominent wind directions from PowerRose
"""
from .mean_wind_dirs import MeanWindDirections
from .prominent_wind_dirs import ProminentWindDirections, row_col_indices
